
// Type-secure C++ interface for multi-dimensional arrays.


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// template class `Section_': Adds a data field to `DAD'.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<int R, class T>
class Section_ : public DAD {
public :

  Section_(const Group& _group, T* _data) :
               DAD(R, _group), data(_data) {}

  T* dat() const {return data ; }

  operator IOArg() const {return IOArg((DAD*) this, data, sizeof(T)) ; }
      // Conversion operator to support `aprintf' etc.

protected :
  Section_(const Group& _group) :
               DAD(R, _group) {}
                   // Used by `Array' subclasses.

  Section_() {}  // Used by `SectionR()'

  T* data ;

} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// template classes `SectionR': Add subscripting operations to         //
//                              `Section_'.                            //
//                                                                     //
// template classes `ArrayR': Add element allocation to `SectionR'     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Section0 : public Section_<0, T> {
public :
  // Constructors

  Section0(const Section0<T>& a) : Section_<0, T>(a.grp(), a.dat()) {

  }

  Section0(const Group& p,
           T* _data) : Section_<0, T>(p, _data) {

  }

  // Simple subscripting

  T& operator()() const {
    return dat() [0] ;
  }

  // Section subscripting

  Section0<T> sect() const {
    return Section0<T>(grp(),
                       dat()) ;
  }

protected :
  Section0() {}  // Used by subclasses.

  Section0(const Group& p) : Section_<0, T>(p) {
      // Used by Array classes to initialize virtual base.

  }

} ;


template<class T>
class Array0 : public Section0<T> {
public:

  Array0(const Group& p = apg) : Section0<T>(p) {
    
    int size = 1 ;

    data = new T [size] ;
  }

  ~Array0() {
    delete [] data ;
  }

private :
  Array0(const Array0& a) : Section0(a) {}  // Prohibit copying.

} ;

template<class T>
class Section1 : public Section_<1, T> {
public :
  // Constructors

  Section1(const Section1<T>& a) : Section_<1, T>(a.grp(), a.dat()) {
    rngs [0] = a.rng(0) ;

    strs [0] = a.jmp(0) ;

    DAD :: ranges  = rngs ;
    DAD :: strides = strs ;
  }

  Section1(const Range& x_0, const Group& p,
           const int str_0, T* _data) : Section_<1, T>(p, _data) {
    rngs [0] = x_0 ;

    strs [0] = str_0 ;

    DAD :: ranges  = rngs ;
    DAD :: strides = strs ;
  }

  // Simple subscripting

  T& operator()(const Location& i_0) const {
    return dat() [str(0).offset(i_0)] ;
  }

  // Section subscripting

  Section1<T> sect(const Trip& i_0) const {
    return Section1<T>(rng(0)(i_0), grp(),
                       jmp(0), dat()) ;
  }

  Section0<T> sect(const Location& i_0) const {
    return Section0<T>(grp() / i_0,
                       dat() + str(0).offset(i_0)) ;
  }

protected :
  Section1() {}  // Used by subclasses.

  Section1(const Group& p) : Section_<1, T>(p) {
      // Used by Array classes to initialize virtual base.

    DAD :: ranges  = rngs ;
    DAD :: strides = strs ;
  }

  Range rngs [1] ;
  int   strs [1] ;
} ;


template<class T>
class Array1 : public Section1<T> {
public:

  Array1(const Range& x_0, const Group& p = apg) : Section1<T>(p) {

    rngs [0] = x_0 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= x_0.volume() ;

    data = new T [size] ;
  }

  ~Array1() {
    delete [] data ;
  }

private :
  Array1(const Array1& a) : Section1(a) {}  // Prohibit copying.

} ;

template<class T>
class Section1S : public virtual Section1<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  Section1S(const Section1S<T>& a) : Section1<T>(a) {}

  Section1S(const Range& x_0, const Group& p,
            const int str_0, T* _data) : Section1<T>(x_0, p,
                                                     str_0, _data) {}
                                                     
  // Simple subscripting

  T& operator()(const int i_0) const {
    return dat() [str(0).offset(rng(0)(i_0))] ;
  }

  // Section subscripting

  Section1S<T> sect(const Trip& i_0) const {
    return Section1S<T>(rng(0)(i_0), grp(),
                        jmp(0), dat()) ;
  }

  Section0<T> sect(const int i_0) const {
    return Section0<T>(grp(),
                       dat() + str(0).offset(rng(0)(i_0))) ;
  }

protected :
  Section1S() {}  // Used by subclasses.
} ;


template<class T>
class Array1S : public Section1S<T>, private virtual Section1<T> {
public:

  Array1S(const int n_0, const Group& p = apg) : Section1<T>(p) {

    rngs [0] = n_0 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= n_0 ;

    data = new T [size] ;
  }

  ~Array1S() {
    delete [] data ;
  }

private :
  Array1S(const Array1S& a) : Section1S(a) {}  // Prohibit copying.

} ;

template<class T>
class Section2 : public Section_<2, T> {
public :
  // Constructors

  Section2(const Section2<T>& a) : Section_<2, T>(a.grp(), a.dat()) {
    rngs [0] = a.rng(0) ;
    rngs [1] = a.rng(1) ;

    strs [0] = a.jmp(0) ;
    strs [1] = a.jmp(1) ;

    DAD :: ranges  = rngs ;
    DAD :: strides = strs ;
  }

  Section2(const Range& x_0,
           const Range& x_1, const Group& p,
           const int str_0,
           const int str_1, T* _data) : Section_<2, T>(p, _data) {
    rngs [0] = x_0 ;
    rngs [1] = x_1 ;

    strs [0] = str_0 ;
    strs [1] = str_1 ;

    DAD :: ranges  = rngs ;
    DAD :: strides = strs ;
  }

  // Simple subscripting

  T& operator()(const Location& i_0,
                const Location& i_1) const {
    return dat() [str(0).offset(i_0) +
                  str(1).offset(i_1)] ;
  }

  // Section subscripting

  Section2<T> sect(const Trip& i_0,
                   const Trip& i_1) const {
    return Section2<T>(rng(0)(i_0),
                       rng(1)(i_1), grp(),
                       jmp(0),
                       jmp(1), dat()) ;
  }

  Section1<T> sect(const Location& i_0,
                   const Trip& i_1) const {
    return Section1<T>(rng(1)(i_1), grp() / i_0,
                       jmp(1), dat() + str(0).offset(i_0)) ;
  }

  Section1<T> sect(const Trip& i_0,
                   const Location& i_1) const {
    return Section1<T>(rng(0)(i_0), grp() / i_1,
                       jmp(0), dat() + str(1).offset(i_1)) ;
  }

  Section0<T> sect(const Location& i_0,
                   const Location& i_1) const {
    return Section0<T>(grp() / i_0 / i_1,
                       dat() + str(0).offset(i_0)
                             + str(1).offset(i_1)) ;
  }

protected :
  Section2() {}  // Used by subclasses.

  Section2(const Group& p) : Section_<2, T>(p) {
      // Used by Array classes to initialize virtual base.

    DAD :: ranges  = rngs ;
    DAD :: strides = strs ;
  }

  Range rngs [2] ;
  int   strs [2] ;
} ;


template<class T>
class Array2 : public Section2<T> {
public:

  Array2(const Range& x_0,
         const Range& x_1, const Group& p = apg) : Section2<T>(p) {

    rngs [0] = x_0 ;
    rngs [1] = x_1 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= x_0.volume() ;
    strs [1] = size ;
    size *= x_1.volume() ;

    data = new T [size] ;
  }

  ~Array2() {
    delete [] data ;
  }

private :
  Array2(const Array2& a) : Section2(a) {}  // Prohibit copying.

} ;

template<class T>
class Section2SD : public virtual Section2<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  Section2SD(const Section2SD<T>& a) : Section2<T>(a) {}

  Section2SD(const Range& x_0,
             const Range& x_1, const Group& p,
             const int str_0,
             const int str_1, T* _data) : Section2<T>(x_0,
                                                      x_1, p,
                                                      str_0,
                                                      str_1, _data) {}
                                                      
  // Simple subscripting

  T& operator()(const int i_0,
                const Location& i_1) const {
    return dat() [str(0).offset(rng(0)(i_0)) +
                  str(1).offset(i_1)] ;
  }

  // Section subscripting

  Section2SD<T> sect(const Trip& i_0,
                     const Trip& i_1) const {
    return Section2SD<T>(rng(0)(i_0),
                         rng(1)(i_1), grp(),
                         jmp(0),
                         jmp(1), dat()) ;
  }

  Section1<T> sect(const int i_0,
                   const Trip& i_1) const {
    return Section1<T>(rng(1)(i_1), grp(),
                       jmp(1), dat() + str(0).offset(rng(0)(i_0))) ;
  }

  Section1S<T> sect(const Trip& i_0,
                    const Location& i_1) const {
    return Section1S<T>(rng(0)(i_0), grp() / i_1,
                        jmp(0), dat() + str(1).offset(i_1)) ;
  }

  Section0<T> sect(const int i_0,
                   const Location& i_1) const {
    return Section0<T>(grp() / i_1,
                       dat() + str(0).offset(rng(0)(i_0))
                             + str(1).offset(i_1)) ;
  }

protected :
  Section2SD() {}  // Used by subclasses.
} ;


template<class T>
class Array2SD : public Section2SD<T>, private virtual Section2<T> {
public:

  Array2SD(const int n_0,
           const Range& x_1, const Group& p = apg) : Section2<T>(p) {

    rngs [0] = n_0 ;
    rngs [1] = x_1 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= n_0 ;
    strs [1] = size ;
    size *= x_1.volume() ;

    data = new T [size] ;
  }

  ~Array2SD() {
    delete [] data ;
  }

private :
  Array2SD(const Array2SD& a) : Section2SD(a) {}  // Prohibit copying.

} ;

template<class T>
class Section2DS : public virtual Section2<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  Section2DS(const Section2DS<T>& a) : Section2<T>(a) {}

  Section2DS(const Range& x_0,
             const Range& x_1, const Group& p,
             const int str_0,
             const int str_1, T* _data) : Section2<T>(x_0,
                                                      x_1, p,
                                                      str_0,
                                                      str_1, _data) {}
                                                      
  // Simple subscripting

  T& operator()(const Location& i_0,
                const int i_1) const {
    return dat() [str(0).offset(i_0) +
                  str(1).offset(rng(1)(i_1))] ;
  }

  // Section subscripting

  Section2DS<T> sect(const Trip& i_0,
                     const Trip& i_1) const {
    return Section2DS<T>(rng(0)(i_0),
                         rng(1)(i_1), grp(),
                         jmp(0),
                         jmp(1), dat()) ;
  }

  Section1S<T> sect(const Location& i_0,
                    const Trip& i_1) const {
    return Section1S<T>(rng(1)(i_1), grp() / i_0,
                        jmp(1), dat() + str(0).offset(i_0)) ;
  }

  Section1<T> sect(const Trip& i_0,
                   const int i_1) const {
    return Section1<T>(rng(0)(i_0), grp(),
                       jmp(0), dat() + str(1).offset(rng(1)(i_1))) ;
  }

  Section0<T> sect(const Location& i_0,
                   const int i_1) const {
    return Section0<T>(grp() / i_0,
                       dat() + str(0).offset(i_0)
                             + str(1).offset(rng(1)(i_1))) ;
  }

protected :
  Section2DS() {}  // Used by subclasses.
} ;


template<class T>
class Array2DS : public Section2DS<T>, private virtual Section2<T> {
public:

  Array2DS(const Range& x_0,
           const int n_1, const Group& p = apg) : Section2<T>(p) {

    rngs [0] = x_0 ;
    rngs [1] = n_1 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= x_0.volume() ;
    strs [1] = size ;
    size *= n_1 ;

    data = new T [size] ;
  }

  ~Array2DS() {
    delete [] data ;
  }

private :
  Array2DS(const Array2DS& a) : Section2DS(a) {}  // Prohibit copying.

} ;

template<class T>
class Section2SS : private virtual Section2<T>,
                   public Section2DS<T>,
                   public Section2SD<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  Section2SS(const Section2SS<T>& a) : Section2<T>(a) {}

  Section2SS(const Range& x_0,
             const Range& x_1, const Group& p,
             const int str_0,
             const int str_1, T* _data) : Section2<T>(x_0,
                                                      x_1, p,
                                                      str_0,
                                                      str_1, _data) {}
                                                      
  // Simple subscripting

  T& operator()(const int i_0,
                const int i_1) const {
    return dat() [str(0).offset(rng(0)(i_0)) +
                  str(1).offset(rng(1)(i_1))] ;
  }

  // Section subscripting

  Section2SS<T> sect(const Trip& i_0,
                     const Trip& i_1) const {
    return Section2SS<T>(rng(0)(i_0),
                         rng(1)(i_1), grp(),
                         jmp(0),
                         jmp(1), dat()) ;
  }

  Section1S<T> sect(const int i_0,
                    const Trip& i_1) const {
    return Section1S<T>(rng(1)(i_1), grp(),
                        jmp(1), dat() + str(0).offset(rng(0)(i_0))) ;
  }

  Section1S<T> sect(const Trip& i_0,
                    const int i_1) const {
    return Section1S<T>(rng(0)(i_0), grp(),
                        jmp(0), dat() + str(1).offset(rng(1)(i_1))) ;
  }

  Section0<T> sect(const int i_0,
                   const int i_1) const {
    return Section0<T>(grp(),
                       dat() + str(0).offset(rng(0)(i_0))
                             + str(1).offset(rng(1)(i_1))) ;
  }

protected :
  Section2SS() {}  // Used by subclasses.
} ;


template<class T>
class Array2SS : public Section2SS<T>, private virtual Section2<T> {
public:

  Array2SS(const int n_0,
           const int n_1, const Group& p = apg) : Section2<T>(p) {

    rngs [0] = n_0 ;
    rngs [1] = n_1 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= n_0 ;
    strs [1] = size ;
    size *= n_1 ;

    data = new T [size] ;
  }

  ~Array2SS() {
    delete [] data ;
  }

private :
  Array2SS(const Array2SS& a) : Section2SS(a) {}  // Prohibit copying.

} ;

template<class T>
class Section3 : public Section_<3, T> {
public :
  // Constructors

  Section3(const Section3<T>& a) : Section_<3, T>(a.grp(), a.dat()) {
    rngs [0] = a.rng(0) ;
    rngs [1] = a.rng(1) ;
    rngs [2] = a.rng(2) ;

    strs [0] = a.jmp(0) ;
    strs [1] = a.jmp(1) ;
    strs [2] = a.jmp(2) ;

    DAD :: ranges  = rngs ;
    DAD :: strides = strs ;
  }

  Section3(const Range& x_0,
           const Range& x_1,
           const Range& x_2, const Group& p,
           const int str_0,
           const int str_1,
           const int str_2, T* _data) : Section_<3, T>(p, _data) {
    rngs [0] = x_0 ;
    rngs [1] = x_1 ;
    rngs [2] = x_2 ;

    strs [0] = str_0 ;
    strs [1] = str_1 ;
    strs [2] = str_2 ;

    DAD :: ranges  = rngs ;
    DAD :: strides = strs ;
  }

  // Simple subscripting

  T& operator()(const Location& i_0,
                const Location& i_1,
                const Location& i_2) const {
    return dat() [str(0).offset(i_0) +
                  str(1).offset(i_1) +
                  str(2).offset(i_2)] ;
  }

  // Section subscripting

  Section3<T> sect(const Trip& i_0,
                   const Trip& i_1,
                   const Trip& i_2) const {
    return Section3<T>(rng(0)(i_0),
                       rng(1)(i_1),
                       rng(2)(i_2), grp(),
                       jmp(0),
                       jmp(1),
                       jmp(2), dat()) ;
  }

  Section2<T> sect(const Location& i_0,
                   const Trip& i_1,
                   const Trip& i_2) const {
    return Section2<T>(rng(1)(i_1),
                       rng(2)(i_2), grp() / i_0,
                       jmp(1),
                       jmp(2), dat() + str(0).offset(i_0)) ;
  }

  Section2<T> sect(const Trip& i_0,
                   const Location& i_1,
                   const Trip& i_2) const {
    return Section2<T>(rng(0)(i_0),
                       rng(2)(i_2), grp() / i_1,
                       jmp(0),
                       jmp(2), dat() + str(1).offset(i_1)) ;
  }

  Section1<T> sect(const Location& i_0,
                   const Location& i_1,
                   const Trip& i_2) const {
    return Section1<T>(rng(2)(i_2), grp() / i_0 / i_1,
                       jmp(2), dat() + str(0).offset(i_0)
                                     + str(1).offset(i_1)) ;
  }

  Section2<T> sect(const Trip& i_0,
                   const Trip& i_1,
                   const Location& i_2) const {
    return Section2<T>(rng(0)(i_0),
                       rng(1)(i_1), grp() / i_2,
                       jmp(0),
                       jmp(1), dat() + str(2).offset(i_2)) ;
  }

  Section1<T> sect(const Location& i_0,
                   const Trip& i_1,
                   const Location& i_2) const {
    return Section1<T>(rng(1)(i_1), grp() / i_0 / i_2,
                       jmp(1), dat() + str(0).offset(i_0)
                                     + str(2).offset(i_2)) ;
  }

  Section1<T> sect(const Trip& i_0,
                   const Location& i_1,
                   const Location& i_2) const {
    return Section1<T>(rng(0)(i_0), grp() / i_1 / i_2,
                       jmp(0), dat() + str(1).offset(i_1)
                                     + str(2).offset(i_2)) ;
  }

  Section0<T> sect(const Location& i_0,
                   const Location& i_1,
                   const Location& i_2) const {
    return Section0<T>(grp() / i_0 / i_1 / i_2,
                       dat() + str(0).offset(i_0)
                             + str(1).offset(i_1)
                             + str(2).offset(i_2)) ;
  }

protected :
  Section3() {}  // Used by subclasses.

  Section3(const Group& p) : Section_<3, T>(p) {
      // Used by Array classes to initialize virtual base.

    DAD :: ranges  = rngs ;
    DAD :: strides = strs ;
  }

  Range rngs [3] ;
  int   strs [3] ;
} ;


template<class T>
class Array3 : public Section3<T> {
public:

  Array3(const Range& x_0,
         const Range& x_1,
         const Range& x_2, const Group& p = apg) : Section3<T>(p) {

    rngs [0] = x_0 ;
    rngs [1] = x_1 ;
    rngs [2] = x_2 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= x_0.volume() ;
    strs [1] = size ;
    size *= x_1.volume() ;
    strs [2] = size ;
    size *= x_2.volume() ;

    data = new T [size] ;
  }

  ~Array3() {
    delete [] data ;
  }

private :
  Array3(const Array3& a) : Section3(a) {}  // Prohibit copying.

} ;

template<class T>
class Section3SDD : public virtual Section3<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  Section3SDD(const Section3SDD<T>& a) : Section3<T>(a) {}

  Section3SDD(const Range& x_0,
              const Range& x_1,
              const Range& x_2, const Group& p,
              const int str_0,
              const int str_1,
              const int str_2, T* _data) : Section3<T>(x_0,
                                                       x_1,
                                                       x_2, p,
                                                       str_0,
                                                       str_1,
                                                       str_2, _data) {}
                                                       
  // Simple subscripting

  T& operator()(const int i_0,
                const Location& i_1,
                const Location& i_2) const {
    return dat() [str(0).offset(rng(0)(i_0)) +
                  str(1).offset(i_1) +
                  str(2).offset(i_2)] ;
  }

  // Section subscripting

  Section3SDD<T> sect(const Trip& i_0,
                      const Trip& i_1,
                      const Trip& i_2) const {
    return Section3SDD<T>(rng(0)(i_0),
                          rng(1)(i_1),
                          rng(2)(i_2), grp(),
                          jmp(0),
                          jmp(1),
                          jmp(2), dat()) ;
  }

  Section2<T> sect(const int i_0,
                   const Trip& i_1,
                   const Trip& i_2) const {
    return Section2<T>(rng(1)(i_1),
                       rng(2)(i_2), grp(),
                       jmp(1),
                       jmp(2), dat() + str(0).offset(rng(0)(i_0))) ;
  }

  Section2SD<T> sect(const Trip& i_0,
                     const Location& i_1,
                     const Trip& i_2) const {
    return Section2SD<T>(rng(0)(i_0),
                         rng(2)(i_2), grp() / i_1,
                         jmp(0),
                         jmp(2), dat() + str(1).offset(i_1)) ;
  }

  Section1<T> sect(const int i_0,
                   const Location& i_1,
                   const Trip& i_2) const {
    return Section1<T>(rng(2)(i_2), grp() / i_1,
                       jmp(2), dat() + str(0).offset(rng(0)(i_0))
                                     + str(1).offset(i_1)) ;
  }

  Section2SD<T> sect(const Trip& i_0,
                     const Trip& i_1,
                     const Location& i_2) const {
    return Section2SD<T>(rng(0)(i_0),
                         rng(1)(i_1), grp() / i_2,
                         jmp(0),
                         jmp(1), dat() + str(2).offset(i_2)) ;
  }

  Section1<T> sect(const int i_0,
                   const Trip& i_1,
                   const Location& i_2) const {
    return Section1<T>(rng(1)(i_1), grp() / i_2,
                       jmp(1), dat() + str(0).offset(rng(0)(i_0))
                                     + str(2).offset(i_2)) ;
  }

  Section1S<T> sect(const Trip& i_0,
                    const Location& i_1,
                    const Location& i_2) const {
    return Section1S<T>(rng(0)(i_0), grp() / i_1 / i_2,
                        jmp(0), dat() + str(1).offset(i_1)
                                      + str(2).offset(i_2)) ;
  }

  Section0<T> sect(const int i_0,
                   const Location& i_1,
                   const Location& i_2) const {
    return Section0<T>(grp() / i_1 / i_2,
                       dat() + str(0).offset(rng(0)(i_0))
                             + str(1).offset(i_1)
                             + str(2).offset(i_2)) ;
  }

protected :
  Section3SDD() {}  // Used by subclasses.
} ;


template<class T>
class Array3SDD : public Section3SDD<T>, private virtual Section3<T> {
public:

  Array3SDD(const int n_0,
            const Range& x_1,
            const Range& x_2, const Group& p = apg) : Section3<T>(p) {

    rngs [0] = n_0 ;
    rngs [1] = x_1 ;
    rngs [2] = x_2 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= n_0 ;
    strs [1] = size ;
    size *= x_1.volume() ;
    strs [2] = size ;
    size *= x_2.volume() ;

    data = new T [size] ;
  }

  ~Array3SDD() {
    delete [] data ;
  }

private :
  Array3SDD(const Array3SDD& a) : Section3SDD(a) {}  // Prohibit copying.

} ;

template<class T>
class Section3DSD : public virtual Section3<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  Section3DSD(const Section3DSD<T>& a) : Section3<T>(a) {}

  Section3DSD(const Range& x_0,
              const Range& x_1,
              const Range& x_2, const Group& p,
              const int str_0,
              const int str_1,
              const int str_2, T* _data) : Section3<T>(x_0,
                                                       x_1,
                                                       x_2, p,
                                                       str_0,
                                                       str_1,
                                                       str_2, _data) {}
                                                       
  // Simple subscripting

  T& operator()(const Location& i_0,
                const int i_1,
                const Location& i_2) const {
    return dat() [str(0).offset(i_0) +
                  str(1).offset(rng(1)(i_1)) +
                  str(2).offset(i_2)] ;
  }

  // Section subscripting

  Section3DSD<T> sect(const Trip& i_0,
                      const Trip& i_1,
                      const Trip& i_2) const {
    return Section3DSD<T>(rng(0)(i_0),
                          rng(1)(i_1),
                          rng(2)(i_2), grp(),
                          jmp(0),
                          jmp(1),
                          jmp(2), dat()) ;
  }

  Section2SD<T> sect(const Location& i_0,
                     const Trip& i_1,
                     const Trip& i_2) const {
    return Section2SD<T>(rng(1)(i_1),
                         rng(2)(i_2), grp() / i_0,
                         jmp(1),
                         jmp(2), dat() + str(0).offset(i_0)) ;
  }

  Section2<T> sect(const Trip& i_0,
                   const int i_1,
                   const Trip& i_2) const {
    return Section2<T>(rng(0)(i_0),
                       rng(2)(i_2), grp(),
                       jmp(0),
                       jmp(2), dat() + str(1).offset(rng(1)(i_1))) ;
  }

  Section1<T> sect(const Location& i_0,
                   const int i_1,
                   const Trip& i_2) const {
    return Section1<T>(rng(2)(i_2), grp() / i_0,
                       jmp(2), dat() + str(0).offset(i_0)
                                     + str(1).offset(rng(1)(i_1))) ;
  }

  Section2DS<T> sect(const Trip& i_0,
                     const Trip& i_1,
                     const Location& i_2) const {
    return Section2DS<T>(rng(0)(i_0),
                         rng(1)(i_1), grp() / i_2,
                         jmp(0),
                         jmp(1), dat() + str(2).offset(i_2)) ;
  }

  Section1S<T> sect(const Location& i_0,
                    const Trip& i_1,
                    const Location& i_2) const {
    return Section1S<T>(rng(1)(i_1), grp() / i_0 / i_2,
                        jmp(1), dat() + str(0).offset(i_0)
                                      + str(2).offset(i_2)) ;
  }

  Section1<T> sect(const Trip& i_0,
                   const int i_1,
                   const Location& i_2) const {
    return Section1<T>(rng(0)(i_0), grp() / i_2,
                       jmp(0), dat() + str(1).offset(rng(1)(i_1))
                                     + str(2).offset(i_2)) ;
  }

  Section0<T> sect(const Location& i_0,
                   const int i_1,
                   const Location& i_2) const {
    return Section0<T>(grp() / i_0 / i_2,
                       dat() + str(0).offset(i_0)
                             + str(1).offset(rng(1)(i_1))
                             + str(2).offset(i_2)) ;
  }

protected :
  Section3DSD() {}  // Used by subclasses.
} ;


template<class T>
class Array3DSD : public Section3DSD<T>, private virtual Section3<T> {
public:

  Array3DSD(const Range& x_0,
            const int n_1,
            const Range& x_2, const Group& p = apg) : Section3<T>(p) {

    rngs [0] = x_0 ;
    rngs [1] = n_1 ;
    rngs [2] = x_2 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= x_0.volume() ;
    strs [1] = size ;
    size *= n_1 ;
    strs [2] = size ;
    size *= x_2.volume() ;

    data = new T [size] ;
  }

  ~Array3DSD() {
    delete [] data ;
  }

private :
  Array3DSD(const Array3DSD& a) : Section3DSD(a) {}  // Prohibit copying.

} ;

template<class T>
class Section3SSD : private virtual Section3<T>,
                    public Section3DSD<T>,
                    public Section3SDD<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  Section3SSD(const Section3SSD<T>& a) : Section3<T>(a) {}

  Section3SSD(const Range& x_0,
              const Range& x_1,
              const Range& x_2, const Group& p,
              const int str_0,
              const int str_1,
              const int str_2, T* _data) : Section3<T>(x_0,
                                                       x_1,
                                                       x_2, p,
                                                       str_0,
                                                       str_1,
                                                       str_2, _data) {}
                                                       
  // Simple subscripting

  T& operator()(const int i_0,
                const int i_1,
                const Location& i_2) const {
    return dat() [str(0).offset(rng(0)(i_0)) +
                  str(1).offset(rng(1)(i_1)) +
                  str(2).offset(i_2)] ;
  }

  // Section subscripting

  Section3SSD<T> sect(const Trip& i_0,
                      const Trip& i_1,
                      const Trip& i_2) const {
    return Section3SSD<T>(rng(0)(i_0),
                          rng(1)(i_1),
                          rng(2)(i_2), grp(),
                          jmp(0),
                          jmp(1),
                          jmp(2), dat()) ;
  }

  Section2SD<T> sect(const int i_0,
                     const Trip& i_1,
                     const Trip& i_2) const {
    return Section2SD<T>(rng(1)(i_1),
                         rng(2)(i_2), grp(),
                         jmp(1),
                         jmp(2), dat() + str(0).offset(rng(0)(i_0))) ;
  }

  Section2SD<T> sect(const Trip& i_0,
                     const int i_1,
                     const Trip& i_2) const {
    return Section2SD<T>(rng(0)(i_0),
                         rng(2)(i_2), grp(),
                         jmp(0),
                         jmp(2), dat() + str(1).offset(rng(1)(i_1))) ;
  }

  Section1<T> sect(const int i_0,
                   const int i_1,
                   const Trip& i_2) const {
    return Section1<T>(rng(2)(i_2), grp(),
                       jmp(2), dat() + str(0).offset(rng(0)(i_0))
                                     + str(1).offset(rng(1)(i_1))) ;
  }

  Section2SS<T> sect(const Trip& i_0,
                     const Trip& i_1,
                     const Location& i_2) const {
    return Section2SS<T>(rng(0)(i_0),
                         rng(1)(i_1), grp() / i_2,
                         jmp(0),
                         jmp(1), dat() + str(2).offset(i_2)) ;
  }

  Section1S<T> sect(const int i_0,
                    const Trip& i_1,
                    const Location& i_2) const {
    return Section1S<T>(rng(1)(i_1), grp() / i_2,
                        jmp(1), dat() + str(0).offset(rng(0)(i_0))
                                      + str(2).offset(i_2)) ;
  }

  Section1S<T> sect(const Trip& i_0,
                    const int i_1,
                    const Location& i_2) const {
    return Section1S<T>(rng(0)(i_0), grp() / i_2,
                        jmp(0), dat() + str(1).offset(rng(1)(i_1))
                                      + str(2).offset(i_2)) ;
  }

  Section0<T> sect(const int i_0,
                   const int i_1,
                   const Location& i_2) const {
    return Section0<T>(grp() / i_2,
                       dat() + str(0).offset(rng(0)(i_0))
                             + str(1).offset(rng(1)(i_1))
                             + str(2).offset(i_2)) ;
  }

protected :
  Section3SSD() {}  // Used by subclasses.
} ;


template<class T>
class Array3SSD : public Section3SSD<T>, private virtual Section3<T> {
public:

  Array3SSD(const int n_0,
            const int n_1,
            const Range& x_2, const Group& p = apg) : Section3<T>(p) {

    rngs [0] = n_0 ;
    rngs [1] = n_1 ;
    rngs [2] = x_2 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= n_0 ;
    strs [1] = size ;
    size *= n_1 ;
    strs [2] = size ;
    size *= x_2.volume() ;

    data = new T [size] ;
  }

  ~Array3SSD() {
    delete [] data ;
  }

private :
  Array3SSD(const Array3SSD& a) : Section3SSD(a) {}  // Prohibit copying.

} ;

template<class T>
class Section3DDS : public virtual Section3<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  Section3DDS(const Section3DDS<T>& a) : Section3<T>(a) {}

  Section3DDS(const Range& x_0,
              const Range& x_1,
              const Range& x_2, const Group& p,
              const int str_0,
              const int str_1,
              const int str_2, T* _data) : Section3<T>(x_0,
                                                       x_1,
                                                       x_2, p,
                                                       str_0,
                                                       str_1,
                                                       str_2, _data) {}
                                                       
  // Simple subscripting

  T& operator()(const Location& i_0,
                const Location& i_1,
                const int i_2) const {
    return dat() [str(0).offset(i_0) +
                  str(1).offset(i_1) +
                  str(2).offset(rng(2)(i_2))] ;
  }

  // Section subscripting

  Section3DDS<T> sect(const Trip& i_0,
                      const Trip& i_1,
                      const Trip& i_2) const {
    return Section3DDS<T>(rng(0)(i_0),
                          rng(1)(i_1),
                          rng(2)(i_2), grp(),
                          jmp(0),
                          jmp(1),
                          jmp(2), dat()) ;
  }

  Section2DS<T> sect(const Location& i_0,
                     const Trip& i_1,
                     const Trip& i_2) const {
    return Section2DS<T>(rng(1)(i_1),
                         rng(2)(i_2), grp() / i_0,
                         jmp(1),
                         jmp(2), dat() + str(0).offset(i_0)) ;
  }

  Section2DS<T> sect(const Trip& i_0,
                     const Location& i_1,
                     const Trip& i_2) const {
    return Section2DS<T>(rng(0)(i_0),
                         rng(2)(i_2), grp() / i_1,
                         jmp(0),
                         jmp(2), dat() + str(1).offset(i_1)) ;
  }

  Section1S<T> sect(const Location& i_0,
                    const Location& i_1,
                    const Trip& i_2) const {
    return Section1S<T>(rng(2)(i_2), grp() / i_0 / i_1,
                        jmp(2), dat() + str(0).offset(i_0)
                                      + str(1).offset(i_1)) ;
  }

  Section2<T> sect(const Trip& i_0,
                   const Trip& i_1,
                   const int i_2) const {
    return Section2<T>(rng(0)(i_0),
                       rng(1)(i_1), grp(),
                       jmp(0),
                       jmp(1), dat() + str(2).offset(rng(2)(i_2))) ;
  }

  Section1<T> sect(const Location& i_0,
                   const Trip& i_1,
                   const int i_2) const {
    return Section1<T>(rng(1)(i_1), grp() / i_0,
                       jmp(1), dat() + str(0).offset(i_0)
                                     + str(2).offset(rng(2)(i_2))) ;
  }

  Section1<T> sect(const Trip& i_0,
                   const Location& i_1,
                   const int i_2) const {
    return Section1<T>(rng(0)(i_0), grp() / i_1,
                       jmp(0), dat() + str(1).offset(i_1)
                                     + str(2).offset(rng(2)(i_2))) ;
  }

  Section0<T> sect(const Location& i_0,
                   const Location& i_1,
                   const int i_2) const {
    return Section0<T>(grp() / i_0 / i_1,
                       dat() + str(0).offset(i_0)
                             + str(1).offset(i_1)
                             + str(2).offset(rng(2)(i_2))) ;
  }

protected :
  Section3DDS() {}  // Used by subclasses.
} ;


template<class T>
class Array3DDS : public Section3DDS<T>, private virtual Section3<T> {
public:

  Array3DDS(const Range& x_0,
            const Range& x_1,
            const int n_2, const Group& p = apg) : Section3<T>(p) {

    rngs [0] = x_0 ;
    rngs [1] = x_1 ;
    rngs [2] = n_2 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= x_0.volume() ;
    strs [1] = size ;
    size *= x_1.volume() ;
    strs [2] = size ;
    size *= n_2 ;

    data = new T [size] ;
  }

  ~Array3DDS() {
    delete [] data ;
  }

private :
  Array3DDS(const Array3DDS& a) : Section3DDS(a) {}  // Prohibit copying.

} ;

template<class T>
class Section3SDS : private virtual Section3<T>,
                    public Section3DDS<T>,
                    public Section3SDD<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  Section3SDS(const Section3SDS<T>& a) : Section3<T>(a) {}

  Section3SDS(const Range& x_0,
              const Range& x_1,
              const Range& x_2, const Group& p,
              const int str_0,
              const int str_1,
              const int str_2, T* _data) : Section3<T>(x_0,
                                                       x_1,
                                                       x_2, p,
                                                       str_0,
                                                       str_1,
                                                       str_2, _data) {}
                                                       
  // Simple subscripting

  T& operator()(const int i_0,
                const Location& i_1,
                const int i_2) const {
    return dat() [str(0).offset(rng(0)(i_0)) +
                  str(1).offset(i_1) +
                  str(2).offset(rng(2)(i_2))] ;
  }

  // Section subscripting

  Section3SDS<T> sect(const Trip& i_0,
                      const Trip& i_1,
                      const Trip& i_2) const {
    return Section3SDS<T>(rng(0)(i_0),
                          rng(1)(i_1),
                          rng(2)(i_2), grp(),
                          jmp(0),
                          jmp(1),
                          jmp(2), dat()) ;
  }

  Section2DS<T> sect(const int i_0,
                     const Trip& i_1,
                     const Trip& i_2) const {
    return Section2DS<T>(rng(1)(i_1),
                         rng(2)(i_2), grp(),
                         jmp(1),
                         jmp(2), dat() + str(0).offset(rng(0)(i_0))) ;
  }

  Section2SS<T> sect(const Trip& i_0,
                     const Location& i_1,
                     const Trip& i_2) const {
    return Section2SS<T>(rng(0)(i_0),
                         rng(2)(i_2), grp() / i_1,
                         jmp(0),
                         jmp(2), dat() + str(1).offset(i_1)) ;
  }

  Section1S<T> sect(const int i_0,
                    const Location& i_1,
                    const Trip& i_2) const {
    return Section1S<T>(rng(2)(i_2), grp() / i_1,
                        jmp(2), dat() + str(0).offset(rng(0)(i_0))
                                      + str(1).offset(i_1)) ;
  }

  Section2SD<T> sect(const Trip& i_0,
                     const Trip& i_1,
                     const int i_2) const {
    return Section2SD<T>(rng(0)(i_0),
                         rng(1)(i_1), grp(),
                         jmp(0),
                         jmp(1), dat() + str(2).offset(rng(2)(i_2))) ;
  }

  Section1<T> sect(const int i_0,
                   const Trip& i_1,
                   const int i_2) const {
    return Section1<T>(rng(1)(i_1), grp(),
                       jmp(1), dat() + str(0).offset(rng(0)(i_0))
                                     + str(2).offset(rng(2)(i_2))) ;
  }

  Section1S<T> sect(const Trip& i_0,
                    const Location& i_1,
                    const int i_2) const {
    return Section1S<T>(rng(0)(i_0), grp() / i_1,
                        jmp(0), dat() + str(1).offset(i_1)
                                      + str(2).offset(rng(2)(i_2))) ;
  }

  Section0<T> sect(const int i_0,
                   const Location& i_1,
                   const int i_2) const {
    return Section0<T>(grp() / i_1,
                       dat() + str(0).offset(rng(0)(i_0))
                             + str(1).offset(i_1)
                             + str(2).offset(rng(2)(i_2))) ;
  }

protected :
  Section3SDS() {}  // Used by subclasses.
} ;


template<class T>
class Array3SDS : public Section3SDS<T>, private virtual Section3<T> {
public:

  Array3SDS(const int n_0,
            const Range& x_1,
            const int n_2, const Group& p = apg) : Section3<T>(p) {

    rngs [0] = n_0 ;
    rngs [1] = x_1 ;
    rngs [2] = n_2 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= n_0 ;
    strs [1] = size ;
    size *= x_1.volume() ;
    strs [2] = size ;
    size *= n_2 ;

    data = new T [size] ;
  }

  ~Array3SDS() {
    delete [] data ;
  }

private :
  Array3SDS(const Array3SDS& a) : Section3SDS(a) {}  // Prohibit copying.

} ;

template<class T>
class Section3DSS : private virtual Section3<T>,
                    public Section3DDS<T>,
                    public Section3DSD<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  Section3DSS(const Section3DSS<T>& a) : Section3<T>(a) {}

  Section3DSS(const Range& x_0,
              const Range& x_1,
              const Range& x_2, const Group& p,
              const int str_0,
              const int str_1,
              const int str_2, T* _data) : Section3<T>(x_0,
                                                       x_1,
                                                       x_2, p,
                                                       str_0,
                                                       str_1,
                                                       str_2, _data) {}
                                                       
  // Simple subscripting

  T& operator()(const Location& i_0,
                const int i_1,
                const int i_2) const {
    return dat() [str(0).offset(i_0) +
                  str(1).offset(rng(1)(i_1)) +
                  str(2).offset(rng(2)(i_2))] ;
  }

  // Section subscripting

  Section3DSS<T> sect(const Trip& i_0,
                      const Trip& i_1,
                      const Trip& i_2) const {
    return Section3DSS<T>(rng(0)(i_0),
                          rng(1)(i_1),
                          rng(2)(i_2), grp(),
                          jmp(0),
                          jmp(1),
                          jmp(2), dat()) ;
  }

  Section2SS<T> sect(const Location& i_0,
                     const Trip& i_1,
                     const Trip& i_2) const {
    return Section2SS<T>(rng(1)(i_1),
                         rng(2)(i_2), grp() / i_0,
                         jmp(1),
                         jmp(2), dat() + str(0).offset(i_0)) ;
  }

  Section2DS<T> sect(const Trip& i_0,
                     const int i_1,
                     const Trip& i_2) const {
    return Section2DS<T>(rng(0)(i_0),
                         rng(2)(i_2), grp(),
                         jmp(0),
                         jmp(2), dat() + str(1).offset(rng(1)(i_1))) ;
  }

  Section1S<T> sect(const Location& i_0,
                    const int i_1,
                    const Trip& i_2) const {
    return Section1S<T>(rng(2)(i_2), grp() / i_0,
                        jmp(2), dat() + str(0).offset(i_0)
                                      + str(1).offset(rng(1)(i_1))) ;
  }

  Section2DS<T> sect(const Trip& i_0,
                     const Trip& i_1,
                     const int i_2) const {
    return Section2DS<T>(rng(0)(i_0),
                         rng(1)(i_1), grp(),
                         jmp(0),
                         jmp(1), dat() + str(2).offset(rng(2)(i_2))) ;
  }

  Section1S<T> sect(const Location& i_0,
                    const Trip& i_1,
                    const int i_2) const {
    return Section1S<T>(rng(1)(i_1), grp() / i_0,
                        jmp(1), dat() + str(0).offset(i_0)
                                      + str(2).offset(rng(2)(i_2))) ;
  }

  Section1<T> sect(const Trip& i_0,
                   const int i_1,
                   const int i_2) const {
    return Section1<T>(rng(0)(i_0), grp(),
                       jmp(0), dat() + str(1).offset(rng(1)(i_1))
                                     + str(2).offset(rng(2)(i_2))) ;
  }

  Section0<T> sect(const Location& i_0,
                   const int i_1,
                   const int i_2) const {
    return Section0<T>(grp() / i_0,
                       dat() + str(0).offset(i_0)
                             + str(1).offset(rng(1)(i_1))
                             + str(2).offset(rng(2)(i_2))) ;
  }

protected :
  Section3DSS() {}  // Used by subclasses.
} ;


template<class T>
class Array3DSS : public Section3DSS<T>, private virtual Section3<T> {
public:

  Array3DSS(const Range& x_0,
            const int n_1,
            const int n_2, const Group& p = apg) : Section3<T>(p) {

    rngs [0] = x_0 ;
    rngs [1] = n_1 ;
    rngs [2] = n_2 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= x_0.volume() ;
    strs [1] = size ;
    size *= n_1 ;
    strs [2] = size ;
    size *= n_2 ;

    data = new T [size] ;
  }

  ~Array3DSS() {
    delete [] data ;
  }

private :
  Array3DSS(const Array3DSS& a) : Section3DSS(a) {}  // Prohibit copying.

} ;

template<class T>
class Section3SSS : private virtual Section3<T>,
                    public Section3DSS<T>,
                    public Section3SDS<T>,
                    public Section3SSD<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  Section3SSS(const Section3SSS<T>& a) : Section3<T>(a) {}

  Section3SSS(const Range& x_0,
              const Range& x_1,
              const Range& x_2, const Group& p,
              const int str_0,
              const int str_1,
              const int str_2, T* _data) : Section3<T>(x_0,
                                                       x_1,
                                                       x_2, p,
                                                       str_0,
                                                       str_1,
                                                       str_2, _data) {}
                                                       
  // Simple subscripting

  T& operator()(const int i_0,
                const int i_1,
                const int i_2) const {
    return dat() [str(0).offset(rng(0)(i_0)) +
                  str(1).offset(rng(1)(i_1)) +
                  str(2).offset(rng(2)(i_2))] ;
  }

  // Section subscripting

  Section3SSS<T> sect(const Trip& i_0,
                      const Trip& i_1,
                      const Trip& i_2) const {
    return Section3SSS<T>(rng(0)(i_0),
                          rng(1)(i_1),
                          rng(2)(i_2), grp(),
                          jmp(0),
                          jmp(1),
                          jmp(2), dat()) ;
  }

  Section2SS<T> sect(const int i_0,
                     const Trip& i_1,
                     const Trip& i_2) const {
    return Section2SS<T>(rng(1)(i_1),
                         rng(2)(i_2), grp(),
                         jmp(1),
                         jmp(2), dat() + str(0).offset(rng(0)(i_0))) ;
  }

  Section2SS<T> sect(const Trip& i_0,
                     const int i_1,
                     const Trip& i_2) const {
    return Section2SS<T>(rng(0)(i_0),
                         rng(2)(i_2), grp(),
                         jmp(0),
                         jmp(2), dat() + str(1).offset(rng(1)(i_1))) ;
  }

  Section1S<T> sect(const int i_0,
                    const int i_1,
                    const Trip& i_2) const {
    return Section1S<T>(rng(2)(i_2), grp(),
                        jmp(2), dat() + str(0).offset(rng(0)(i_0))
                                      + str(1).offset(rng(1)(i_1))) ;
  }

  Section2SS<T> sect(const Trip& i_0,
                     const Trip& i_1,
                     const int i_2) const {
    return Section2SS<T>(rng(0)(i_0),
                         rng(1)(i_1), grp(),
                         jmp(0),
                         jmp(1), dat() + str(2).offset(rng(2)(i_2))) ;
  }

  Section1S<T> sect(const int i_0,
                    const Trip& i_1,
                    const int i_2) const {
    return Section1S<T>(rng(1)(i_1), grp(),
                        jmp(1), dat() + str(0).offset(rng(0)(i_0))
                                      + str(2).offset(rng(2)(i_2))) ;
  }

  Section1S<T> sect(const Trip& i_0,
                    const int i_1,
                    const int i_2) const {
    return Section1S<T>(rng(0)(i_0), grp(),
                        jmp(0), dat() + str(1).offset(rng(1)(i_1))
                                      + str(2).offset(rng(2)(i_2))) ;
  }

  Section0<T> sect(const int i_0,
                   const int i_1,
                   const int i_2) const {
    return Section0<T>(grp(),
                       dat() + str(0).offset(rng(0)(i_0))
                             + str(1).offset(rng(1)(i_1))
                             + str(2).offset(rng(2)(i_2))) ;
  }

protected :
  Section3SSS() {}  // Used by subclasses.
} ;


template<class T>
class Array3SSS : public Section3SSS<T>, private virtual Section3<T> {
public:

  Array3SSS(const int n_0,
            const int n_1,
            const int n_2, const Group& p = apg) : Section3<T>(p) {

    rngs [0] = n_0 ;
    rngs [1] = n_1 ;
    rngs [2] = n_2 ;
    
    int size = 1 ;
    strs [0] = size ;
    size *= n_0 ;
    strs [1] = size ;
    size *= n_1 ;
    strs [2] = size ;
    size *= n_2 ;

    data = new T [size] ;
  }

  ~Array3SSS() {
    delete [] data ;
  }

private :
  Array3SSS(const Array3SSS& a) : Section3SSS(a) {}  // Prohibit copying.

} ;

